//**************************************
//
// BMP Format - File Format support for ImageExtender class
//
// NOTE: CURRENT CAPABILITIES!!
//		- Load supports:
//					16 bit X1-R5-G5-B5 (without alpha)
//					24 bit    R8-G8-B8 (without alpha)
//					32 bit A8-R8-G8-B8 (with alpha) -- This is currently the only way to load a BMP with alpha
//		- Save supports:                               (this should change soon to allow separate alpha BMP files to be used)
//					Nothing!
//
//**************************************

#include "gamex-image-bmp.hpp"

// Function: LoadBMP
//
//		filename			Name of BMP file to load
//		img					ImageX object to fill with image data from bitmap 

void ImageExt::LoadBMP (char *filename, ImageX *img)
{
	HBITMAP hbmp = (HBITMAP) LoadImage(NULL,filename,IMAGE_BITMAP,0,0,LR_LOADFROMFILE|LR_CREATEDIBSECTION);

	bool ok = false;
	if(hbmp) {
		BITMAP bitmap;
		GetObject(hbmp,sizeof(bitmap),&bitmap);

		if(bitmap.bmBits != NULL) {

			LPBITMAPINFO lpbitmapinfo;
			int size = (bitmap.bmBitsPixel != 8) ? sizeof(BITMAPINFOHEADER) : sizeof(BITMAPINFOHEADER)+256*sizeof(RGBQUAD);
			lpbitmapinfo = (LPBITMAPINFO)malloc(size);
			BITMAPINFO bitmapinfo = *lpbitmapinfo;
			GetObject(hbmp,size,&bitmapinfo);

			if(bitmap.bmWidth > 2048 || bitmap.bmHeight > 2048) {
				char errstr [256];
				sprintf(errstr,"The %dx%d image in file %s is too large for GameX to use.\nNeither width nor height may exceed 2048 pixels.",bitmap.bmWidth,bitmap.bmHeight,filename);
				MessageBox (NULL, errstr, "GameX Error", MB_OK|MB_ICONSTOP);
				if(bitmap.bmWidth > 2048) bitmap.bmWidth = 2048;
				if(bitmap.bmHeight > 2048) bitmap.bmHeight = 2048;
			}

			int pitch = bitmap.bmWidthBytes;

			switch(bitmap.bmBitsPixel)
			{
			case 8: {
	/*			RGBQUAD rgbquadpalette [256];
				PALETTEENTRY palette [256];

	//			for(int i = sizeof(BITMAPINFOHEADER)/sizeof(lpbitmapinfo) ; i < sizeof(BITMAPINFOHEADER)/sizeof(lpbitmapinfo) + 256*sizeof(RGBQUAD)/sizeof(lpbitmapinfo) ; i++) {
				for(int i = 0; i < 256 ; i++) {
					int a = sizeof(BITMAPINFOHEADER)/sizeof(lpbitmapinfo);
					int b = sizeof(RGBQUAD)/sizeof(lpbitmapinfo);
					rgbquadpalette[i] = bitmapinfo.bmiColors[a + i*b];
					palette[i].peBlue = rgbquadpalette[i].rgbBlue;
					palette[i].peGreen = rgbquadpalette[i].rgbGreen;
					palette[i].peRed = rgbquadpalette[i].rgbRed;
					palette[i].peFlags = PC_NOCOLLAPSE;
				}*/
	/*
				if(FAILED(lpddpal->SetEntries(0,0,256,bitmap.palette)))
				{
					int p = 5;
				}
	*/
				img->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_INDEXED);
			}
			case 16: {
				int bmp_mask_red=0, bmp_mask_green=0, bmp_mask_blue=0;
				int bmp_shift_red, bmp_shift_green, bmp_shift_blue;

				BITMAPV4HEADER bitmapinfo4;
				GetObject(hbmp,sizeof(bitmapinfo4),&bitmapinfo4);
	/*
				// get some well-guarded info needed to decode the bitmap:
				// (why isn't this in bitmapinfo.bV4RedMask, bitmapinfo.bV4GreenMask etc.?)
				bmp_mask_red = bitmapinfo4.bV4Endpoints.ciexyzRed.ciexyzY;
				bmp_mask_green = bitmapinfo4.bV4Endpoints.ciexyzRed.ciexyzZ;
				bmp_mask_blue = bitmapinfo4.bV4Endpoints.ciexyzGreen.ciexyzX;
	*/			// just in case
				if(bmp_mask_red   <= 0) bmp_mask_red   = 0x7C00;
				if(bmp_mask_green <= 0) bmp_mask_green = 0x03E0;
				if(bmp_mask_blue  <= 0) bmp_mask_blue  = 0x001F;

				bmp_shift_red   = GameX.GetShift(bmp_mask_red);
				bmp_shift_green = GameX.GetShift(bmp_mask_green);
				bmp_shift_blue  = GameX.GetShift(bmp_mask_blue);

				img->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_HIGHCOLOR);
				XBYTE2 * pData = (XBYTE2 *) img->m_data;
				XBYTE2 * bmpData = (XBYTE2 *) bitmap.bmBits;
				for(int y = 0 ; y < img->m_yres ; y++)
				for(int x = 0 ; x < img->m_xres ; x++) {
						int n = x+y*img->m_xres; // pixel address -- destination
						int yi = img->m_yres-y-1; // inverted y, because bmp's are stored upside-down
						int ni = x+yi*pitch/2; // inverted pixel address -- source
						int clr = bmpData[ni] ;
						int blue  = (clr & bmp_mask_blue) >> bmp_shift_blue;  // decode blue
						int green = (clr & bmp_mask_green) >> (bmp_shift_green-1);  // decode green
						int red   = (clr & bmp_mask_red) >> bmp_shift_red; // decode red
						pData[n] = ((red << 11) + (green << 5) + (blue << 0)); // re-encode
				}
			 } break;
			case 24: {
				img->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_TRUECOLOR);
				XBYTE * pData = (XBYTE *) img->m_data;
				XBYTE * bmpData = (XBYTE *) bitmap.bmBits;
				for(int y = 0 ; y < img->m_yres ; y++)
				for(int x = 0 ; x < 3*img->m_xres ; x+=3) {
					int n = x+y*img->m_xres*3; // pixel address -- destination
					int yi = img->m_yres-y-1; // inverted y, because bmp's are stored upside-down
					int ni = x+yi*pitch; // inverted pixel address -- source
					pData[n+2] = bmpData[ni];   // blue // (not only is it upside-down, but RGB is backwards)
					pData[n+1] = bmpData[ni+1]; // green
					pData[n]   = bmpData[ni+2]; // red
				}
			 } break;
			case 32: {
				img->Size(bitmap.bmWidth, bitmap.bmHeight, IMG_TRUECOLORX|IMG_MERGEALPHA);
				XBYTE * pData = (XBYTE *) img->m_data;
				XBYTE * bmpData = (XBYTE *) bitmap.bmBits;
				for(int y = 0 ; y < img->m_yres ; y++)
				for(int x = 0 ; x < 4*img->m_xres ; x+=4) {
					int n = x+y*img->m_xres*4; // pixel address -- destination
					int yi = img->m_yres-y-1; // inverted y, because bmp's are stored upside-down
					int ni = x+yi*pitch; // inverted pixel address -- source
					pData[n+3] = bmpData[ni+3]; // alpha
					pData[n+2] = bmpData[ni+2]; // red
					pData[n+1] = bmpData[ni+1]; // green
					pData[n]   = bmpData[ni]; // blue
				}


			 } break;
			}
			DeleteObject((HGDIOBJ) hbmp);
			ok = true;
		} else {
			DeleteObject((HGDIOBJ) hbmp);
			ok = false;
		}
	}
	if(!ok) {
		// something went wrong, either the file wasn't found
		// or it's in a nonstandard format such as [A4 R4 G4 B4] or [R5 G6 B5]
		char disp [500];
		#ifdef WINDX_DEBUG
			sprintf (disp, "The image in file %s did not load. \n\nMake sure paths and names of BMP images are correct and that the executable is in the correct location to find the images.\nAlso make sure the BMP is either 16 bit (X1 R5 G5 B5), 24 bit (R8 G8 B8), or 32 bit (A8 R8 G8 B8).\nLastly, if the file is in the right place and in the right format, make sure it isn't already open.", filename); 
			img->Size(64,64,IMG_HIGHCOLOR);
			img->Fill(255,128,0); // if debugging is set, make missing image shows up orange
		#else
			sprintf (disp, "The image in file %s did not load. \n\nMake sure the image is in that location and that this game's executable is in the correct location to find it.", filename); 
		#endif

			MessageBox (NULL, disp, "GameX Error", MB_OK|MB_ICONSTOP);
	}
}

