//
// GameX - ImageX Class Code 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source Liscense. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include <math.h>
#include <string.h>

#include "gamex-image.hpp"

ImageX::ImageX (void)
{
	Reset();
}

ImageX::ImageX (int xr, int yr, ImageOps ops)
{
	Size (xr, yr, ops);
}

ImageX::~ImageX (void)
{
	if(m_surface != NULL) {
		if(GameX.GetDD()) // if we haven't already released all of DirectDraw
			m_surface->Release(); // release the surface data (from video memory)
		m_surface = NULL;
	}
	DeleteData(); // delete the non-surface data
	m_xres = 0; m_yres = 0; m_size = 0;
	m_options = IMG_UNDEF;
}

void ImageX::Reset (void)
{
	m_xres = 0; m_yres = 0; m_size = 0;
	m_options = IMG_UNDEF;
	DeleteData();
	m_usage = DRAW_UNDEF;
	status = IMG_NOTREADY;
	m_hassurface = 0;
	m_filename = NULL;
	m_surface = NULL;
	m_device = NULL;
}

void ImageX::AllocateSurface ()
{
/*	DDSURFACEDESC2 ddsd;	

	if (m_surface != NULL) {m_surface->Release() ;	m_surface = NULL;}

	ZeroMemory (&ddsd, sizeof (ddsd)) ;
	// Set back surface description
	ddsd.dwSize = sizeof (ddsd) ;
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT;
	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue = 0; // specify black as transparent if drawn masked
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
	ddsd.dwWidth = m_xres;
	ddsd.dwHeight = m_yres;

	// Try to create secondary surface 
	if (FAILED(GameX.GetDD()->CreateSurface (&ddsd, &m_surface, NULL))) {
		if(required) { // if we absolutely need to allocate the surface,
			ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY | DDSCAPS_3DDEVICE;
			if (FAILED(GameX.GetDD()->CreateSurface (&ddsd, &m_surface, NULL))) {
				if (m_surface != NULL) {m_surface->Release(); m_surface = NULL;}
				m_hassurface = false; // still failed, nothing more we can do at this point...
			} else
				m_hassurface = true; // success
		}
		else {
			if (m_surface != NULL) {m_surface->Release(); m_surface = NULL;}
			m_hassurface = false; // if we failed, don't try again; there is no speed gain from having it in system memory
		}
	}
	else
		m_hassurface = true; // success
*/
//	LPDIRECTDRAWSURFACE7
	bool textureSurface = !(m_usage & LOAD_NON3D);
	int alphaSurface = ((m_usage & LOAD_ALPHA) && (m_options & IMG_ALPHA) && !(m_usage & LOAD_MASKED)) ? 2 : 0;
	bool targetable = (m_usage & LOAD_TARGETABLE) ? true : false;
	m_surface = GameX.CreateImageSurface(m_xres, m_yres, textureSurface, alphaSurface, targetable);

	DDPIXELFORMAT ddpf;
	ZeroMemory (&ddpf, sizeof(DDPIXELFORMAT)) ;
	ddpf.dwSize = sizeof(ddpf);
	m_surface->GetPixelFormat(&ddpf); // get the pixel format DirectDraw made the surface in,
	m_surf_bpp = ddpf.dwRGBBitCount; // usually this is win_screenbpp or 32 for alpha, but it's not guaranteed

	if(targetable && textureSurface) {
		// this will probably get the best SOFTWARE 3D rendering we have,
		// since the hardware device is likely to be tied up by the back buffer,
		// which usually needs the speed more than anything else.
		// getting software rendering only means that rendering TO this surface is done in software,
		// but rendering FROM this or any surface to the back buffer will be accelerated either way.
		if(GameX.CreateDeviceForSurface(m_device, m_surface)) {

			m_device->SetTextureStageState(0,D3DTSS_ADDRESSU,D3DTADDRESS_CLAMP); // we don't want texture wrapping
			m_device->SetTextureStageState(0,D3DTSS_ADDRESSV,D3DTADDRESS_CLAMP); // that doesn't work right on most image sizes anyway

			m_device->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE); // we have no D3D light sources, for now
			m_device->SetRenderState(D3DRENDERSTATE_COLORVERTEX, TRUE); // we want to allow per-vertex coloring

		}
/*		D3DVIEWPORT7 view;
		view.dwX=0;
		view.dwY=0; 
		view.dwWidth=m_xres;
		view.dwHeight=m_yres;
		view.dvMinZ=0.0f;
		view.dvMaxZ=1.0f;
		m_device->SetViewport(&view);*/
	}

//	if(alphaSurface) {
//		ConvertFormat(IMG_TRUECOLOR | IMG_MERGEALPHA);
//		TurnOn(IMG_MERGEALPHA);
//	}

	if(m_surface == NULL)
		m_hassurface = 0; // failure
	else {
		m_hassurface = (textureSurface) ? 2 : 1; // success: 1 for normal, 2 for 3D accelerated

		DDSURFACEDESC2 ddsd;	
		ZeroMemory (&ddsd, sizeof (ddsd)) ;
		ddsd.dwSize = sizeof (ddsd) ;
		m_surface->GetSurfaceDesc(&ddsd);
		m_surf_width = ddsd.dwWidth;
		m_surf_height = ddsd.dwHeight;

		RECT boundary[1] = {0,0,m_xres,m_yres};
		LPDIRECTDRAWCLIPPER clipper;
		LPRGNDATA rgn_data;
		if (FAILED(GameX.GetDD()->CreateClipper(0,&clipper,NULL))) {
			return;
		}
		int rectSize = sizeof(RECT);
		int rgnHeaderSize = sizeof(RGNDATAHEADER);
		rgn_data = (LPRGNDATA)malloc(rgnHeaderSize+rectSize);
		memcpy(rgn_data->Buffer, boundary, rectSize);
		rgn_data->rdh.dwSize = rgnHeaderSize;
		rgn_data->rdh.nRgnSize = rectSize;
		rgn_data->rdh.nCount = 1;
		rgn_data->rdh.iType = RDH_RECTANGLES;
		rgn_data->rdh.rcBound = boundary[0];
		if (FAILED(clipper->SetClipList(rgn_data,0)) || FAILED(m_surface->SetClipper(clipper))) {
			free(rgn_data);
			return;
		}
		free(rgn_data);
	}
}

void ImageX::DataToSurface (void) // temporary: m_data should eventually be replaced entirely by m_surface
{
	// copy the pixels onto the surface, so that the surface can be blitted for us later
	if(m_hassurface == NULL || m_data == NULL) {
		return;
	}

	DDSURFACEDESC2 ddsd;	
	ZeroMemory (&ddsd, sizeof (ddsd)) ;
	ddsd.dwSize = sizeof (ddsd) ;

 	HRESULT status = m_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL); 	// lock drawing surface
	if(status != DD_OK) {
		GameX.DirectDrawError(status);
//		MessageBox (GameX.GetWindow(), "Error: GAMEX_LOCK_FAILED (Unable to lock offscreen image surface)", "GameX Error", MB_OK|MB_ICONSTOP);
		return;
	}


	if(m_options & IMG_INDEXED) {
		MessageBox(GameX.GetWindow(), "Error: Unsupported transfer operation (8 bit image to surface).", "GameX Error",MB_OK|MB_ICONSTOP) ;
		// not supported yet
	}
	else
	if(m_options & IMG_HIGHCOLOR) {
		if(m_options & IMG_ALPHA) {
			// if this image has a separate alpha channel,
			// combine the alpha and color data into a 32 bit surface texture for D3D to draw
			// (D3D's conversion from 32 bit to 16 bit on each draw can't be helped)
			XBYTE2 *pDataOld;
			XBYTE *pAlphaOld;
			XBYTE4 *pDataNew;
	//		newdata = new XBYTE[m_size * GetBytesPerPixel(ops)];			
			pDataOld = (XBYTE2 *) m_data; 
			pAlphaOld = (XBYTE *) m_alpha; 
			pDataNew = (XBYTE4 *) ddsd.lpSurface;
			int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			int offset = 0;

			for(int y = 0 ; y < m_yres ; y++)
			{
				offset = y * adjusted_pitch;
				for(int x = 0 ; x < m_xres ; x++)
				{
					int clr = *pDataOld;
					int blue  = (clr & win_maskb) >> win_shiftb << 3; // decode blue
					int green = (clr & win_maskg) >> win_shiftg << 2; // decode green
					int red   = (clr & win_maskr) >> win_shiftr << 3; // decode red
					int alpha;
					if(m_alpha == NULL) alpha = 0;
					else alpha = *pAlphaOld;
					int newColor = RGBA_MAKE(red,green,blue,alpha);
					*(pDataNew + offset) = newColor;

					// clamp by one pixel to prevent oddities at scaled edges
					if(x == m_xres-1 && m_surf_width > m_xres) *(pDataNew + offset + 1) = newColor;
					if(y == m_yres-1 && m_surf_height > m_yres) {
						*(pDataNew + offset + adjusted_pitch) = newColor;
						if(x == m_xres-1 && m_surf_width > m_xres) *(pDataNew + offset + adjusted_pitch + 1) = newColor;
					}
					if(x == 0 && m_surf_width > m_xres) *(pDataNew + offset + m_surf_width-1) = newColor;
					if(y == 0 && m_surf_height > m_yres) *(pDataNew + offset + adjusted_pitch*(m_surf_height-1)) = newColor;

					offset++;
					pDataOld++;
					pAlphaOld++;
				}
			}
	/*		XBYTE2 * buf_data = (XBYTE2 *) ddsd.lpSurface;
			int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
		//	int adjusted_pitch = ddsd.lPitch >> (GetBytesPerPixel(m_options)-1); // pitch adjusted for bit depth (>> 1 for 16 bit, etc.)
			int adjusted_pitch = ddsd.lPitch >> 1; // >> 1 for 16 bits
			XBYTE2 *pData = (XBYTE2*) m_data;
			int offset = 0;

			for(int y = 0 ; y < m_yres ; y++)
			{
				for(int x = 0 ; x < m_xres ; x++)
				{
					*(buf_data + offset) = *pData ;
					pData++; offset++;
				}
				offset += adjusted_pitch - m_xres;
			}*/

		} else {

			XBYTE2 * buf_data = (XBYTE2 *) ddsd.lpSurface;
			int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
		//	int adjusted_pitch = ddsd.lPitch >> (GetBytesPerPixel(m_options)-1); // pitch adjusted for bit depth (>> 1 for 16 bit, etc.)
			int adjusted_pitch = ddsd.lPitch / 2; // /2 for 16 bits
			XBYTE2 *pData = (XBYTE2*) m_data;

			int offset = 0;

			for(int y = 0 ; y < m_yres ; y++)
			{
				for(int x = 0 ; x < m_xres ; x++)
				{
					int newColor = *pData;
					*(buf_data + offset) = newColor;

					// clamp by one pixel to prevent oddities at scaled edges
					if(x == m_xres-1 && m_surf_width > m_xres) *(buf_data + offset + 1) = newColor;
					if(y == m_yres-1 && m_surf_height > m_yres) {
						*(buf_data + offset + adjusted_pitch) = newColor;
						if(x == m_xres-1 && m_surf_width > m_xres) *(buf_data + offset + adjusted_pitch + 1) = newColor;
					}
					if(x == 0 && m_surf_width > m_xres) *(buf_data + offset + m_surf_width-1) = newColor;
					if(y == 0 && m_surf_height > m_yres) *(buf_data + offset + adjusted_pitch*(m_surf_height-1)) = newColor;


					pData++;
					offset++;
				}
				offset += adjusted_pitch - m_xres;
			}
		}
	}
	else
	if(m_options & IMG_TRUECOLOR) {
		if(m_options & IMG_ALPHA) {
			// if this image has a separate alpha channel,
			// combine the alpha and color data into a 32 bit surface texture for D3D to draw
			XBYTE *pDataOld;
			XBYTE *pAlphaOld;
			XBYTE4 *pDataNew;
			pDataOld = (XBYTE *) m_data; 
			pAlphaOld = (XBYTE *) m_alpha; 
			pDataNew = (XBYTE4 *) ddsd.lpSurface;
			int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			int offset = 0;

			for(int y = 0 ; y < m_yres ; y++)
			{
				offset = y * adjusted_pitch;
				for(int x = 0 ; x < m_xres ; x++)
				{
					int clr = ((*pDataOld)<<16) | ((*(pDataOld+1))<<8) | ((*(pDataOld+2))<<0);
					int alpha;
					if(m_alpha == NULL) alpha = 0;
					else alpha = *pAlphaOld;
					int newColor = clr + (alpha << 24);
					*(pDataNew + offset) = newColor;

					// clamp by one pixel to prevent oddities at scaled edges
					if(x == m_xres-1 && m_surf_width > m_xres) *(pDataNew + offset + 1) = newColor;
					if(y == m_yres-1 && m_surf_height > m_yres) {
						*(pDataNew + offset + adjusted_pitch) = newColor;
						if(x == m_xres-1 && m_surf_width > m_xres) *(pDataNew + offset + adjusted_pitch + 1) = newColor;
					}
					if(x == 0 && m_surf_width > m_xres) *(pDataNew + offset + m_surf_width-1) = newColor;
					if(y == 0 && m_surf_height > m_yres) *(pDataNew + offset + adjusted_pitch*(m_surf_height-1)) = newColor;

					pDataOld+=3;
					pAlphaOld++;
					offset++;
				}
			}

		} else {
			if(m_surf_bpp == 24) {
				XBYTE * buf_data = (XBYTE *) ddsd.lpSurface;
				int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
				int adjusted_pitch = ddsd.lPitch / 3; // /3 for 24 bits
				XBYTE *pData = (XBYTE*) m_data;
				int offset = 0;

				for(int y = 0 ; y < m_yres ; y++)
				{
					for(int x = 0 ; x < m_xres ; x++)
					{
						int red = *pData;
						int green = *(pData+1);
						int blue = *(pData+2);
						*(buf_data + offset) = red;
						*(buf_data + offset+1) = green;
						*(buf_data + offset+2) = blue;

						// clamp by one pixel to prevent oddities at scaled edges
						if(x == m_xres-1 && m_surf_width > m_xres) {
							*(buf_data + offset + 3) = red;
							*(buf_data + offset + 4) = green;
							*(buf_data + offset + 5) = blue;
						}
						if(y == m_yres-1 && m_surf_height > m_yres) {
							*(buf_data + offset + adjusted_pitch  ) = red;
							*(buf_data + offset + adjusted_pitch+1) = green;
							*(buf_data + offset + adjusted_pitch+2) = blue;
							if(x == m_xres-1 && m_surf_width > m_xres) {
								*(buf_data + offset + adjusted_pitch+3) = red;
								*(buf_data + offset + adjusted_pitch+4) = green;
								*(buf_data + offset + adjusted_pitch+5) = blue;
							}
						}
						if(x == 0 && m_surf_width > m_xres) {
							*(buf_data + offset + (m_surf_width-1)*3  ) = red;
							*(buf_data + offset + (m_surf_width-1)*3+1) = green;
							*(buf_data + offset + (m_surf_width-1)*3+2) = blue;
						}
						if(y == 0 && m_surf_height > m_yres) {
							*(buf_data + offset + adjusted_pitch*(m_surf_height-1)  ) = red;
							*(buf_data + offset + adjusted_pitch*(m_surf_height-1)+1) = green;
							*(buf_data + offset + adjusted_pitch*(m_surf_height-1)+2) = blue;
						}

						pData+=3;
						offset+=3;
					}
					offset += adjusted_pitch - m_xres;
				}
			}
			else
			if(m_surf_bpp == 32) {
				XBYTE4 * buf_data = (XBYTE4 *) ddsd.lpSurface;
				int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
				int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
				XBYTE *pData = (XBYTE*) m_data;
				int offset = 0;

				for(int y = 0 ; y < m_yres ; y++)
				{
					for(int x = 0 ; x < m_xres ; x++)
					{
						int clr = ((*pData)<<16) | ((*(pData+1))<<8) | ((*(pData+2))<<0);
						*(buf_data + offset) = clr;

						// clamp by one pixel to prevent oddities at scaled edges
						if(x == m_xres-1 && m_surf_width > m_xres) {
							*(buf_data + offset + 1) = clr;
						}
						if(y == m_yres-1 && m_surf_height > m_yres) {
							*(buf_data + offset + adjusted_pitch) = clr;
							if(x == m_xres-1 && m_surf_width > m_xres) {
								*(buf_data + offset + adjusted_pitch+1) = clr;
							}
						}
						if(x == 0 && m_surf_width > m_xres) *(buf_data + offset + m_surf_width-1) = clr;
						if(y == 0 && m_surf_height > m_yres) *(buf_data + offset + adjusted_pitch*(m_surf_height-1)) = clr;

						pData+=3;
						offset++;
					}
					offset += adjusted_pitch - m_xres;
				}
			}
		}
	}
	else
	if(m_options & IMG_TRUECOLORX) {
		if(m_options & IMG_ALPHA) {
			// if this image has a separate alpha channel,
			// combine the alpha and color data into a 32 bit surface texture for D3D to draw
			XBYTE4 *pDataOld;
			XBYTE *pAlphaOld;
			XBYTE4 *pDataNew;
			pDataOld = (XBYTE4 *) m_data; 
			pAlphaOld = (XBYTE *) m_alpha; 
			pDataNew = (XBYTE4 *) ddsd.lpSurface;
			int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			int offset = 0;

			for(int y = 0 ; y < m_yres ; y++)
			{
				offset = y * adjusted_pitch;
				for(int x = 0 ; x < m_xres ; x++)
				{
					int clr = *pDataOld;
					int alpha;
					if(m_alpha == NULL) alpha = 0; // this should be 0 if the alpha is already in the color
					else alpha = *pAlphaOld;
					int newColor = clr + (alpha << 24);
					*(pDataNew + offset) = newColor;

					// clamp by one pixel to prevent oddities at scaled edges
					if(x == m_xres-1 && m_surf_width > m_xres) *(pDataNew + offset + 1) = newColor;
					if(y == m_yres-1 && m_surf_height > m_yres) {
						*(pDataNew + offset + adjusted_pitch) = newColor;
						if(x == m_xres-1 && m_surf_width > m_xres) *(pDataNew + offset + adjusted_pitch + 1) = newColor;
					}
					if(x == 0 && m_surf_width > m_xres) *(pDataNew + offset + m_surf_width-1) = newColor;
					if(y == 0 && m_surf_height > m_yres) *(pDataNew + offset + adjusted_pitch*(m_surf_height-1)) = newColor;

					pDataOld++;
					pAlphaOld++;
					offset++;
				}
			}
		} else {
			XBYTE4 * buf_data = (XBYTE4 *) ddsd.lpSurface;
			int bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;
			int adjusted_pitch = ddsd.lPitch / 4; // /4 for 32 bits
			XBYTE4 *pData = (XBYTE4*) m_data;
			int offset = 0;

			for(int y = 0 ; y < m_yres ; y++)
			{
				for(int x = 0 ; x < m_xres ; x++)
				{
					int newColor = *pData;
					*(buf_data + offset) = newColor;

					// clamp by one pixel to prevent oddities at scaled edges
					if(x == m_xres-1 && m_surf_width > m_xres) *(buf_data + offset + 1) = newColor;
					if(y == m_yres-1 && m_surf_height > m_yres) {
						*(buf_data + offset + adjusted_pitch) = newColor;
						if(x == m_xres-1 && m_surf_width > m_xres) *(buf_data + offset + adjusted_pitch + 1) = newColor;
					}
					if(x == 0 && m_surf_width > m_xres) *(buf_data + offset + m_surf_width-1) = newColor;
					if(y == 0 && m_surf_height > m_yres) *(buf_data + offset + adjusted_pitch*(m_surf_height-1)) = newColor;

					pData++;
					offset++;
				}
				offset += adjusted_pitch - m_xres;
			}
		}
	}

 	status = m_surface->Unlock(NULL); // unlock drawing surface
	if(status != DD_OK){ 	
		GameX.DirectDrawError(status);
//		MessageBox (GameX.GetWindow(), "Error: GAMEX_UNLOCK_FAILED (Unable to unlock offscreen image surface)", "GameX Error", MB_OK|MB_ICONSTOP);
		return;
	}

	if(GameX.CanDeleteMData()) {
		if(!(m_usage & LOAD_NON3D) || !((m_usage & LOAD_ALPHA && !(m_usage & LOAD_MASKED)))) {
			if(m_data != NULL) {
				DeleteData();
			}
		}
	}
}

inline int ImageX::GetBytesPerPixel (ImageOps ops)
{												// Pixel Format
	if (ops & IMG_MERGEALPHA) {					// I R G B A
		if (ops & IMG_INDEXED) return 2;		// 8-0-0-0-8
		if (ops & IMG_LOWCOLOR) return 2;		// 0-4-4-4-4
		if (ops & IMG_HIGHCOLOR) return 3;		// 0-6-5-5-8
		if (ops & IMG_TRUECOLOR) return 4;		// 0-8-8-8-8
		if (ops & IMG_TRUECOLORX) return 4;		// 0-8-8-8-8
	} else {
		if (ops & IMG_INDEXED) return 1;		// 8-0-0-0-0
		if (ops & IMG_LOWCOLOR) return IMG_UNDEF;// undefined
		if (ops & IMG_HIGHCOLOR) return 2;		// 0-6-5-5-0
		if (ops & IMG_TRUECOLOR) return 3;		// 0-8-8-8-0
		if (ops & IMG_TRUECOLORX) return 4;		// 0-8-8-8-8
	}
	return IMG_UNDEF;
}

inline int ImageX::GetFormat (ImageOps ops)
{
	if (ops & IMG_INDEXED) return IMG_INDEXED;
	if (ops & IMG_LOWCOLOR) return IMG_LOWCOLOR;
	if (ops & IMG_HIGHCOLOR) return IMG_HIGHCOLOR;
	if (ops & IMG_TRUECOLOR) return IMG_TRUECOLOR;
	if (ops & IMG_TRUECOLORX) return IMG_TRUECOLORX;
	return IMG_UNDEF;
}

inline int ImageX::GetAlpha (ImageOps ops)
{	
	if (ops & IMG_ALPHA) return IMG_ALPHA;
	return IMG_UNDEF;
}

inline void ImageX::TurnOff (ImageOps opt)
{	
	m_options ^= opt;
}

inline void ImageX::TurnOn (ImageOps opt)
{	
	m_options |= opt;
}

void ImageX::Size (int xr, int yr, ImageOps ops)
{
	if(GetFormat(ops) == IMG_UNDEF) {
#ifdef _DEBUG
		MessageBox(GameX.GetWindow(), "Error: Image options passed to ImageX::Size() must include a format such as IMG_HIGHCOLOR", "GameX Error",MB_OK|MB_ICONSTOP) ;
#endif
		return;
	}

	int bpp = GetBytesPerPixel (ops);
	
	if (bpp != -1) {
		if (m_data!=NULL) {delete [] m_data; m_data = NULL;}	
		m_data = new XBYTE[xr * yr * bpp];
	}
	if (ops & IMG_ALPHA) {
		if (m_alpha!=NULL) {delete [] m_alpha; m_alpha = NULL;}
		m_alpha = new XBYTE[xr * yr];
	}	
	m_xres = xr; m_yres = yr; m_size = xr * yr;
	m_options = ops;
#ifdef WINDX_DEBUG
	Fill (255,0,0); // if debugging is set, make unfilled or improperly filled images show up red
#else
	Fill (0,0,0);
#endif
}

void ImageX::Create (int xr, int yr, ImageUsage what)
{	
	switch (what) {
		case LOAD_ALPHA: Size (xr, yr, IMG_HIGHCOLOR | IMG_ALPHA); break;	
		default: Size (xr, yr, IMG_HIGHCOLOR); break;
	}
	Fill (0,0,0);
	UseWith(what);
}

void ImageX::Fill (int r, int g, int b) // r, g, and b are out of 255
{
	XBYTE2 *pData;
			
	pData = (XBYTE2*) m_data;
	for (int n=0; n < m_size; n++) {
 		*pData++ = (((r >> 3) << win_shiftr) + ((g >> 2) << win_shiftg) + ((b >> 3) << win_shiftb));
	}
}

void ImageX::UseWith (ImageUsage what)
{
	m_usage = what;
	if(!(m_options & IMG_MERGEALPHA)) {
		switch(GameX.win_screenbpp) {
		case 8:
			ConvertFormat(IMG_INDEXED);
			break;
		case 16:
			ConvertFormat(IMG_HIGHCOLOR);
			break;
		case 24:
			ConvertFormat(IMG_TRUECOLOR);
			break;
		case 32:
			ConvertFormat(IMG_TRUECOLORX);
			break;
		}
	}
	if(m_usage & LOAD_GRAY) {
		ConvertMDataToGrayscale();
	}
	m_size = m_xres*m_yres;

	bool makeSurface = true;
	bool saveData = !GameX.CanDeleteMData();

//	if(what & MASKED)
//		ConvertAlphaToMask() ; // if it has an alpha channel, convert that to on/off transparency
//	if(what & ALPHA || what & MASKBLEND || what & CUSTOM || what & ROTATABLE)
//		saveData = true; // disallow deletion of non-surface data
	if(m_options & IMG_ALPHA) {
		if(what & LOAD_MASKED)
			ConvertAlphaToMask();					   			
///		else
///			ConvertAlphaToBlend();					   			
	} else {
		if(m_options & IMG_MERGEALPHA)
			m_options |= IMG_ALPHA;
		else
		if(what & LOAD_ALPHA)
			if(!(what & LOAD_MASKED))
				ConvertMaskToAlpha();
	}
	if(!((m_usage & LOAD_ALPHA) && (m_options & IMG_ALPHA) && !(m_usage & LOAD_MASKED))) {
		if(m_options & IMG_ALPHA) {
			if(m_options & IMG_MERGEALPHA) {
				TurnOff(IMG_MERGEALPHA);
				if(GameX.win_screenbpp == 8) ConvertFormat(IMG_INDEXED);
				if(GameX.win_screenbpp == 16) ConvertFormat(IMG_HIGHCOLOR);
				if(GameX.win_screenbpp == 24) ConvertFormat(IMG_TRUECOLOR);
			}
			TurnOff(IMG_ALPHA);
		}
		if (m_alpha!=NULL) {delete [] m_alpha; m_alpha = NULL;}
	}
	if(makeSurface) {
		AllocateSurface(); // if image is small enough to be worth storing as a surface
		DataToSurface();
		if(!saveData)
			DeleteData();
	}
	
	status = IMG_OK;
}

void ImageX::ConvertMDataToGrayscale (void)
{
	switch (GetFormat(m_options)) {
		case IMG_INDEXED: {
			MessageBox(GameX.GetWindow(), "Error: Unsupported conversion operation (8 bit image to grayscale).", "GameX Error",MB_OK|MB_ICONSTOP) ;
		} break;
		case IMG_HIGHCOLOR: {
			XBYTE2 *pData = (XBYTE2*) m_data;
			for (int n=0; n<m_size; n++) {				
				int red = (*pData & win_maskr) >> win_shiftr;
				int green = (*pData & win_maskg) >> win_shiftg;
				int blue = (*pData & win_maskb) >> win_shiftb;
				int grey = (red+green+blue)/3;
				*pData++ = (grey << win_shiftr) | (grey << win_shiftg) | (grey << win_shiftb);
			}
		} break;
		case IMG_TRUECOLOR: {			
			XBYTE *pData = (XBYTE*) m_data;
			for (int n=0; n<m_size; n++) {	// Alpha channel is gray of primary data
				int blue = *pData;
				int green = *(pData+1);
				int red = *(pData+2);
				int grey = (red+green+blue)/3;
				*pData++ = grey; // blue->grey
				*pData++ = grey; // green->grey
				*pData++ = grey; // red->grey
			}
		} break;
		case IMG_TRUECOLORX: {			
			XBYTE *pData = (XBYTE*) m_data;
			for (int n=0; n<m_size; n++) {	// Alpha channel is gray of primary data
				int blue = *pData;
				int green = *(pData+1);
				int red = *(pData+2);
				int grey = (red+green+blue)/3;
				*pData++ = grey; // blue->grey
				*pData++ = grey; // green->grey
				*pData++ = grey; // red->grey
				*pData++; // leave alpha or X alone
			}
		} break;
	}
}

void ImageX::ConvertMaskToAlpha (void)
{
	// main reason for doing this:
	// D3D renders color key edges in a very ugly way,
	// but it handles alpha channels just fine (esp. with rotation).
	// maybe not worth it though,
	// only takes effect when alpha is specified for an image with no alpha channel

	if (GetAlpha(m_options)!=IMG_ALPHA) {
		if (m_alpha!=NULL) {delete [] m_alpha; m_alpha = NULL;}
		m_alpha = new XBYTE[m_xres * m_yres];
		XBYTE *pAlphaData;
		XBYTE2 *pData;
		pData = (XBYTE2*) m_data;
		pAlphaData = m_alpha;
		for (int n=0; n<m_size; n++) {
			if(*pData == 0) *pAlphaData = 255;
			else			*pAlphaData = 0;
			*pAlphaData++;
			*pData++;
		}
		m_options |= IMG_ALPHA;
//		m_options |= IMG_MASKED;
	}
}

void ImageX::SetPixel (int x, int y, int r, int g, int b)
{
	if (x >=0 && y>=0 && x < m_xres && y < m_yres) {
		XBYTE2 *pData = (XBYTE2*) m_data;	
		pData += (y*m_xres)+x;
		*pData = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
	}
}

void ImageX::GetPixel (int x, int y, int &r, int &g, int &b)
{
	if (x >=0 && y>=0 && x < m_xres && y < m_yres) {
		XBYTE2 *pData = (XBYTE2*) m_data;
		pData += (y*m_xres)+x;
		r = (*pData & win_maskr) >> win_shiftr;
		g = (*pData & win_maskg) >> win_shiftg;
		b = (*pData & win_maskb) >> win_shiftb;	
	} else {
		r = 0; g = 0; b = 0;
	}
}

void ImageX::ConvertFormat (ImageOps ops)
{	
	XBYTE * newdata;
	
	switch (GetFormat(m_options)) {			// Source Format:
		case IMG_INDEXED:					//   Indexed Color (8 bit)
			switch (GetFormat(ops))	{	// Target Format:
				case IMG_HIGHCOLOR:	//   High Color (16 bit)
					MessageBox(GameX.GetWindow(), "Error: Unsupported color conversion operation (8 to 16 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break; // conversion from 8 bit color not yet supported
				case IMG_TRUECOLOR:	//   True Color (24 bit)
					MessageBox(GameX.GetWindow(), "Error: Unsupported color conversion operation (8 to 24 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break; // conversion from 8 bit color not yet supported
				case IMG_TRUECOLORX:	//   True Color with X channel (32 bit)
					MessageBox(GameX.GetWindow(), "Error: Unsupported color conversion operation (8 to 32 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break; // conversion from 8 bit color not yet supported
			}
			break;
		case IMG_HIGHCOLOR:					//   High Color (16 bit)
			switch (GetFormat(ops))	{	// Target Format:
				case IMG_INDEXED:	//   High Color (8 bit)
					MessageBox(GameX.GetWindow(), "Error: Unsupported color conversion operation (16 to 8 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break; // not yet supported
				case IMG_TRUECOLOR:	{ //   True Color (24 bit)
//					MessageBox(GameX.GetWindow(), "Error: Unsupported color conversion operation (16 to 24 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					XBYTE2 *pDataOld;
					XBYTE *pDataNew;
					newdata = new XBYTE[m_size * GetBytesPerPixel(ops)];			
					pDataOld = (XBYTE2 *) m_data; 
					pDataNew = (XBYTE *) newdata;

					if(!win_shiftr)
						GameX.ReportProblem(GAMEX_NO_INIT) ;

					for (int n=0; n<m_size; n++) {
						int clr = *pDataOld++;
						int blue  = (((clr & 0x001F) >> 0 ) << 3);
						int green = (((clr & 0x07E0) >> 5 ) << 2);
						int red   = (((clr & 0xF800) >> 11) << 3);
						*pDataNew++ = red;
						*pDataNew++ = green;
						*pDataNew++ = blue;
					}
					delete [] m_data; m_data = NULL;
					m_data = newdata;			
					break;
				}
				case IMG_TRUECOLORX: {	//   True Color with X channel (32 bit)
					XBYTE2 *pDataOld;
					XBYTE4 *pDataNew;
					newdata = new XBYTE[m_size * GetBytesPerPixel(ops)];			
					pDataOld = (XBYTE2 *) m_data; 
					pDataNew = (XBYTE4 *) newdata;

					if(!win_shiftr)
						GameX.ReportProblem(GAMEX_NO_INIT) ;

					for (int n=0; n<m_size; n++) {
						int clr = *pDataOld++;
						int blue  = (((clr & 0x001F) >> 0 ) << 3);
						int green = (((clr & 0x07E0) >> 5 ) << 2);
						int red   = (((clr & 0xF800) >> 11) << 3);
						*pDataNew++ = (red << win_shiftr) + (green << win_shiftg) + (blue << win_shiftb);
					}
					delete [] m_data; m_data = NULL;
					m_data = newdata;			
					break;
				}
			}
			break;							// Source Format:
		case IMG_TRUECOLOR:					//   True Color (24 bit)
			switch (GetFormat(ops))	{			// Target Format:
				case IMG_INDEXED:			//   Indexed Color (8 bit)
					MessageBox(GameX.GetWindow(), "Error: Unsupported color conversion operation (24 to 8 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break; // not yet supported
				case IMG_HIGHCOLOR:	{			//   High Color (16 bit)
					XBYTE *pDataOld;
					XBYTE2 *pDataNew;
					newdata = new XBYTE[m_size * GetBytesPerPixel(ops)];			
					pDataOld = (XBYTE *) m_data; 
					pDataNew = (XBYTE2 *) newdata;

					if(!win_shiftr) {
						GameX.ReportProblem(GAMEX_NO_INIT) ;
		//				MessageBox(GameX.GetWindow(), "Warning: You seem to be loading an image before GameX has been initialized. Make sure you have no statically activated constructors that call imageExt.Load().", "GameX Warning",MB_OK|MB_ICONEXCLAMATION) ;
					}
		//				win_shiftr = 11; win_shiftg = 6; win_shiftb = 0;

					for (int n=0; n<m_size; n++) {
						*pDataNew = (*pDataOld++ >> 3) << win_shiftr;
						*pDataNew |= (*pDataOld++ >> 2) << win_shiftg;
						*pDataNew++ |= (*pDataOld++ >> 3) << win_shiftb;
					}
					delete [] m_data; m_data = NULL;
					m_data = newdata;			
					break;
				}
				case IMG_TRUECOLORX: {	//   True Color with X channel (32 bit)
					XBYTE *pDataOld;
					XBYTE4 *pDataNew;
					newdata = new XBYTE[m_size * GetBytesPerPixel(ops)];			
					pDataOld = (XBYTE *) m_data; 
					pDataNew = (XBYTE4 *) newdata;

					if(!win_shiftr)
						GameX.ReportProblem(GAMEX_NO_INIT) ;

					for (int n=0; n<m_size; n++) {
						int newPixel = 0;
						newPixel |= *pDataOld++ << win_shiftr;
						newPixel |= *pDataOld++ << win_shiftg;
						newPixel |= *pDataOld++ << win_shiftb;
						*pDataNew++ = newPixel;
					}
					delete [] m_data; m_data = NULL;
					m_data = newdata;
					break;
				}
			}
			break;							// Source Format:
		case IMG_TRUECOLORX:				//   True Color with X channel (32 bit)
			switch (GetFormat(ops))	{		// Target Format:
				case IMG_INDEXED:			//   Indexed Color (8 bit)
					MessageBox(GameX.GetWindow(), "Error: Unsupported color conversion operation (32 to 8 bit).", "GameX Error",MB_OK|MB_ICONSTOP) ;
					break; // conversion to 8-bit color not yet supported
				case IMG_HIGHCOLOR: {			//   High Color (16 bit)
					XBYTE4 *pDataOld;
					XBYTE2 *pDataNew;
					newdata = new XBYTE[m_size * GetBytesPerPixel(ops)];			
					pDataOld = (XBYTE4 *) m_data; 
					pDataNew = (XBYTE2 *) newdata;

					if(!win_shiftr)
						GameX.ReportProblem(GAMEX_NO_INIT) ;

					for (int n=0; n<m_size; n++) {
						int oldPixel = *pDataOld++;
						int red = ((oldPixel & 0x00FF0000) >> 16);
						int green = ((oldPixel & 0x0000FF00) >> 8);
						int blue = ((oldPixel & 0x000000FF) >> 0);
						int newPixel = ((red >> 3) << 11) | ((green >> 2) << 5) | ((blue >> 3) << 0);
						*pDataNew++ = newPixel;
					}
					delete [] m_data; m_data = NULL;
					m_data = newdata;
					break;
				}
				case IMG_TRUECOLOR: {			//   True Color (24 bit)
					XBYTE4 *pDataOld;
					XBYTE *pDataNew;
					newdata = new XBYTE[m_size * GetBytesPerPixel(ops)];			
					pDataOld = (XBYTE4 *) m_data; 
					pDataNew = (XBYTE *) newdata;

					if(!win_shiftr)
						GameX.ReportProblem(GAMEX_NO_INIT) ;

					for (int n=0; n<m_size; n++) {
						int oldPixel = *pDataOld++;
						int red = ((oldPixel & 0x00FF0000) >> 16);
						int green = ((oldPixel & 0x0000FF00) >> 8);
						int blue = ((oldPixel & 0x000000FF) >> 0);
						*pDataNew++ = red;
						*pDataNew++ = green;
						*pDataNew++ = blue;
					}
					delete [] m_data; m_data = NULL;
					m_data = newdata;
					break;
				}
			}
			break;
	}
	TurnOff (GetFormat(m_options));
//	TurnOn (GetFormat(ops));
	TurnOn (ops);
}

void ImageX::ConvertAlphaToMask (void)
{
	// NOTE: IF NO ALPHA CHANNEL IS LOADED, any pixels with value 0 (totally black)
	// in the primary data with be treated as transparent. DrawMasked
	// will draw all 0 value pixels transparent. This is the desired behavior.
	// If an alpha channel >is< present, 0 value pixels will be mapped to value 1,
	// and the alpha channel will be used to determine which pixels to make transparent
	// by setting them to 0.
	if (GetAlpha(m_options)==IMG_ALPHA) {
		switch (GetFormat(m_options)) {
		case IMG_HIGHCOLOR: {
			XBYTE2 *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_size; n++) {
				if (*pData==0) *pData = 1;
				if (*pAlphaData < 16) *pData = 0;
				*pData++;
				*pAlphaData++;
			}
		} break;
		}

//		// if we couldn't possibly want to keep the alpha channel...
//		if(!(m_usage & ALPHA || m_usage & CUSTOM || m_usage & ROTATABLE)) {
//			delete [] m_alpha; m_alpha = NULL; // delete it
			TurnOff (IMG_ALPHA);
			TurnOn (IMG_MASKED);
//		}
	}
}
/*
void ImageX::ConvertAlphaToBlend (void)
{
	if (m_options & IMG_ALPHA) {		
		switch (GetFormat(m_options)) {
		case IMG_TRUECOLOR: {			
			XBYTE *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_size; n++) {
//				*pData++ = (*pData * *pAlphaData)>>8;
//				*pData++ = (*pData * *pAlphaData)>>8;
//				*pData++ = (*pData * *pAlphaData)>>8;
				*pAlphaData++ = 255 - *pAlphaData;
			}
		} break;
		case IMG_HIGHCOLOR: {
//			XBYTE2 r, g, b;
			XBYTE2 *pData;
			XBYTE *pAlphaData;			
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;			
			for (int n=0; n<m_size; n++) {				
//				r = (((*pData & win_maskr) >> win_shiftr) * *pAlphaData) >> 8;
//				g = (((*pData & win_maskg) >> win_shiftg) * *pAlphaData) >> 8;
//				b = (((*pData & win_maskb) >> win_shiftb) * *pAlphaData) >> 8;
//				*pData++ = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
				*pAlphaData = 255 - *pAlphaData;
				pAlphaData++;
			}
		} break;
		}
	} else {										// NO ALPHA CHANNEL LOADED
		if (m_alpha!=NULL) {delete [] m_alpha; m_alpha = NULL;}	
		m_alpha = new XBYTE[m_size];		// Create an alpha channel	
		TurnOn (IMG_ALPHA);		
		switch (GetFormat(m_options)) {
		case IMG_TRUECOLOR: {			
			XBYTE *pData;
			XBYTE *pAlphaData;
			pData = (XBYTE*) m_data;
			pAlphaData = m_alpha;
			for (int n=0; n<m_size; n++) {	// Alpha channel is gray of primary data
				*pAlphaData = (*pData + *(pData+1) + *(pData+2))/3;
//				*pData++ = (*pData * *pAlphaData)>>8;
//				*pData++ = (*pData * *pAlphaData)>>8;
//				*pData++ = (*pData * *pAlphaData)>>8;
				*pAlphaData++ = 255 - *pAlphaData;
			}
		} break;
		case IMG_HIGHCOLOR: {
			XBYTE2 r, g, b;
			XBYTE2 *pData;
			XBYTE *pAlphaData;			
			pData = (XBYTE2*) m_data;
			pAlphaData = m_alpha;			
			for (int n=0; n<m_size; n++) {				
				r = (*pData & win_maskr) >> win_shiftr;
				g = (*pData & win_maskg) >> win_shiftg;
				b = (*pData & win_maskb) >> win_shiftb;
				*pAlphaData = min((r + g + b)*16/3, 255);
				r = (r * *pAlphaData) >> 8;
				g = (g * *pAlphaData) >> 8;
				b = (b * *pAlphaData) >> 8;
//				*pData++ = (r << win_shiftr) + (g << win_shiftg) + (b << win_shiftb);
				*pAlphaData = 255 - *pAlphaData;
				pAlphaData++;
			}
		} break;
		}
	}
}
*/
/*
void ImageX::ScaleAndRotate (ImageX *dest, float ang, float scale, ImageUsage destUsage)
{	
//#ifdef WINDX_DEBUG
//	if(!(m_usage & ROTATABLE)) {
//		MessageBox(GameX.GetWindow(), "ERROR: You must set ROTATABLE for any images that need to be scaled or rotated. As in, IMAGE | ROTATABLE.\nNormally, the ROTATABLE flag can be set when calling Load().\nTo rotate an image that was the destination of a rotation, i.e. if image A is rotated to image B and you want to rotate image B to image C,\n then pass in ROTATABLE as an extra parameter of Scale / Rotate when rotating from image A to B.", "GameX Error",MB_OK|MB_ICONEXCLAMATION) ;
//		return;
//	}
//#endif
	
//	dest->m_usage = (destUsage) ? (destUsage) : (m_usage ^ ROTATABLE);

	dest->m_usage = destUsage ? destUsage : m_usage;
//	if(m_usage & TARGETABLE) destUsage |= TARGETABLE;

//	dest->m_usage = destUsage|m_usage;

	dest->m_filename = NULL;
	if(!dest->m_hassurface)
//		if(dest->m_usage & IMAGE && dest->m_size >= 16*16 || dest->m_usage & MASKED && dest->m_size >= 24*24)
			dest->AllocateSurface () ;
//		else
//			dest->m_hassurface = false;
//	if(dest->m_hassurface && !(dest->m_usage & ROTATABLE) && !saveData) {
//		RotateImage (dest, ang, scale, true);
//	}
//	else {
		RotateImage (dest, ang, scale);
		RotateAlpha (dest, ang, scale);
		dest->status = IMG_OK;
		dest->DataToSurface () ;
//	}
}
		
void ImageX::RotateImage (ImageX *dest, float ang, float scale, bool toVideo)
{
	// NOTE: ASSUMES BOTH IMAGES ARE IMG_HIGHCOLOR

	if (dest!=NULL && m_data!=NULL) {
		XBYTE2 *src_pix;
		XBYTE2 *dest_pix;
		XBYTE2 *dest_start;
		float delt_x, delt_y;
		float start_x, start_y;
		float off_x, off_y;
		float pix_x, pix_y, h;
		int y, cnt;
		int rowSize;

		h = (float) sqrt(dest->m_xres * dest->m_xres + dest->m_yres * dest->m_yres);
		start_x = (m_xres/2.0f) + cosf((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		start_y = (m_yres/2.0f) - sinf((135-ang)*DEGtoRAD)*(h/2.0f)*(m_xres/m_yres)/scale+1; // the +1 fixes a vertical off-by-one bug
		delt_x = cosf(ang*DEGtoRAD)/scale;
		delt_y = sinf(ang*DEGtoRAD)*(m_xres/m_yres)/scale;
		off_x = delt_x * dest->m_xres;
		off_y = delt_y * dest->m_xres;
		if(toVideo) {
			DDSURFACEDESC2 ddsd;	
			ZeroMemory (&ddsd, sizeof (ddsd)) ;
			ddsd.dwSize = sizeof (ddsd) ;

 			HRESULT status = dest->m_surface->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);	// lock drawing surface
			if(status != DD_OK) {
//				MessageBox (GameX.GetWindow(), "Error: GAMEX_LOCK_FAILED (Unable to lock rotating image surface)", "GameX Error", MB_OK|MB_ICONSTOP);
				GameX.DirectDrawError(status);
				return;
			}
			dest_start = (XBYTE2 *) ddsd.lpSurface;
			rowSize = ddsd.lPitch >> 1; // >> 1 for 16 bits
		}
		else {
			dest_start = (XBYTE2*) dest->m_data;
			rowSize = dest->m_xres;
		}
		for (y=0; y<dest->m_yres; y++) {	
			cnt = dest->m_xres;						// Start with full pixel set in line
			pix_x = start_x + off_x-1; //-1 fixes off-by-1 bug				// Goto end of line and scan backwards
			pix_y = start_y + off_y;
			dest_pix = dest_start + cnt-1;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x -= delt_x;					// Scan until end of span found
				pix_y -= delt_y;
				cnt--;
				*dest_pix-- = 0;
			}	
			dest_pix = dest_start;
			pix_x = start_x;						// Goto beginning of line and scan forward
			pix_y = start_y;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x += delt_x;					// Scan until start of span found
				pix_y += delt_y;
				cnt--;
				*dest_pix++ = 0;
			}
			while (cnt>0) {
				src_pix = ((XBYTE2*) m_data) + (int) pix_y*m_xres + (int) pix_x;
				*dest_pix++ = *src_pix;			
				pix_x += delt_x; pix_y += delt_y;
				cnt--;
			}
			dest_start += rowSize;
			start_x -= delt_y;
			start_y += delt_x;
		}
		if(toVideo) {
 			status = dest->m_surface->Unlock(NULL); 	// unlock drawing surface
			if(status != DD_OK){
				GameX.DirectDrawError(status);
//				MessageBox (GameX.GetWindow(), "Error: GAMEX_UNLOCK_FAILED (Unable to unlock offscreen image surface)", "GameX Error", MB_OK|MB_ICONSTOP);
				return;
			}
		}
	} else {
//		if(!this->m_usage) return;
//		GameX.Draw(PLAIN|CENTERED, *this, dest->m_xres/2, dest->m_yres/2, *dest, ang, scale);
	}
}

void ImageX::RotateAlpha (ImageX *dest, float ang, float scale)
{
	if (dest->m_options & IMG_ALPHA && m_alpha!=NULL && dest->m_alpha!=NULL) {
		XBYTE *src_pix;
		XBYTE *dest_pix;
		XBYTE *dest_start;
		float delt_x, delt_y;
		float start_x, start_y;
		float off_x, off_y;
		float pix_x, pix_y, h;
		int y, cnt;

		h = (float) sqrt(dest->m_xres * dest->m_xres + dest->m_yres * dest->m_yres);
		start_x = (m_xres/2.0f) + cosf((135-ang)*DEGtoRAD)*(h/2.0f)/scale;
		start_y = (m_yres/2.0f) - sinf((135-ang)*DEGtoRAD)*(h/2.0f)/scale+1; // the +1 is to work around an off-by-one bug
		delt_x = cosf(ang*DEGtoRAD)/scale;
		delt_y = sinf(ang*DEGtoRAD)/scale;
		off_x = delt_x * dest->m_xres;
		off_y = delt_y * dest->m_xres;	
		dest_start = (XBYTE*) dest->m_alpha;
		for (y=0; y<dest->m_yres; y++) {	
			cnt = dest->m_xres;						// Start with full pixel set in line
			pix_x = start_x + off_x-1; // -1 avoids off-by-one bug		// Goto end of line and scan backwards
			pix_y = start_y + off_y;
			dest_pix = dest_start + cnt-1;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x -= delt_x;					// Scan until end of span found
				pix_y -= delt_y;
				cnt--;
				*dest_pix-- = 255;
			}	
			dest_pix = dest_start;
			pix_x = start_x;						// Goto beginning of line and scan forward
			pix_y = start_y;
			while ((pix_x<0 || pix_x>=m_xres || pix_y<0 || pix_y>=m_yres) && cnt>0) {
				pix_x += delt_x;					// Scan until start of span found
				pix_y += delt_y;
				cnt--;
				*dest_pix++ = 255;
			}
			while (cnt>0) {
				src_pix = ((XBYTE*) m_alpha) + (int) pix_y*m_xres + (int) pix_x;
				*dest_pix++ = *src_pix;			
				pix_x += delt_x; pix_y += delt_y;
				cnt--;
			}
			dest_start += dest->m_xres;
			start_x -= delt_y;
			start_y += delt_x;
		}
	}
}
*/
// to be called on images that have lost their surface and become corrupted
// only seems to happen when interrupting fullscreen mode
// or entering a screensaver or something, 
// and then it only happens somtimes, but consistently when it does
void ImageX::Restore (void)
{
	if(m_device != NULL) { 
		unsigned long temp_long = -1;
		status = m_device->ValidateDevice(&temp_long);

		if(status == DDERR_SURFACELOST) { // in newer versions of DirectX this is really D3DERR_DEVICELOST, which makes more sense
			m_device->Release();
			m_surface->Unlock(NULL); // unlock drawing surface in case it's locked
			bool success = GameX.CreateDeviceForSurface(m_device, m_surface);

#ifdef WINDX_DEBUG
			if(success)
				debug.Output("WinDX: An ImageX's 3D rendering device has been lost and is now properly restored.\n","");
			else
				debug.Output("WinDX: An ImageX's 3D rendering device has been lost and could NOT be restored.\n","");
#endif
		}
	}

//	debug.Output("Checking to see if we need to to restore an ImageX\n","");
	if(m_surface && m_surface->IsLost())
	{
		m_surface->Restore(); // restore the surface (usually doesn't work right)
		if(m_data != NULL) {
			DataToSurface(); // if we have a copy of the image data, copy it over
#ifdef WINDX_DEBUG
			debug.Output("An ImageX was lost and has been restored using a system memory copy.\n","");
#endif
		}
		else if(m_filename != NULL) { // if we don't have the image data anymore,
			m_surface->Release();  // try reloading it from file as a last resort
			m_surface = NULL;
			DeleteData();
			m_xres = 0; m_yres = 0; m_size = 0;
			m_options = IMG_UNDEF;
			ImageExt imageExt;
			imageExt.Load(m_filename,this,m_usage);
			debug.Output("An ImageX was lost and has been restored by re-loading it from file.\n","");
		} else { // otherwise, not much else we can do...
#ifdef WINDX_DEBUG
			debug.Output("An ImageX was lost and may not be fully restored.\n","");
#endif
		}
	}
}


ImageExt::ImageExt ()
{
	m_pImg = NULL;
}

ImageExt::ImageExt (ImageX* pImg)
{
	m_pImg = pImg;
}

ImageExt::~ImageExt ()
{
	m_pImg = NULL;
}

void ImageExt::Load (char *filename, ImageX *img, ImageUsage what)
{
	img->Reset();
	int i = 0, dot = 50000;
	while(filename[i] != 0) {
		if(filename[i] == '.')
			dot = i;
		i++;
	}
	if(i-dot>3 && (filename[dot+1]=='b' || filename[dot+1]=='B') && (filename[dot+2]=='m' || filename[dot+2]=='M') && (filename[dot+3]=='p' || filename[dot+3]=='P'))
		LoadBMP (filename, img);
	else
		LoadTiff (filename, img);
	img->UseWith (what);
	img->m_filename = filename;
}
